import pathlib
import re

import docutils.nodes as nodes
import docutils.parsers.rst as rst
import nbformat.v4 as nbformat
from docutils.parsers.rst.directives.admonitions import BaseAdmonition

import esbonio

RESOURCES = "resources"


class solution(nodes.General, nodes.Element):
    pass


class SolutionDirective(BaseAdmonition):

    NAME = "solution"

    has_content = True
    node_class = solution

    def run(self):
        (soln,) = super().run()
        return [soln]


def visit_solution(self, node):
    self.body.append('<details class="admonition note">\n')
    self.body.append(
        '<summary class="admonition-title"> Solution (click to reveal)</summary>\n'
    )


def depart_solution(self, node):
    self.body.append("</details>\n")


class tutorial(nodes.General, nodes.Element):
    pass


class TutorialDirective(rst.Directive):

    NAME = "tutorial"

    def run(self):
        return [tutorial("")]


def visit_tutorial(self, node):
    pass


def depart_tutorial(self, node):
    pass


def _no_op(self, node):
    pass


no_op = _no_op, _no_op


class NotebookTranslator(nodes.NodeVisitor):
    """Walk an rst doctree and convert it into a Jupyer Notebook."""

    def __init__(self, document):
        super().__init__(document)
        self.cells = []
        self.level = 0
        self.prefix = None

    def asnotebook(self):
        return nbformat.new_notebook(cells=self.cells)

    def astext(self):
        return nbformat.writes(self.asnotebook())

    @property
    def current_cell(self):

        if len(self.cells) == 0:
            return None

        return self.cells[-1]

    def new_cell(self, cell_type: str):
        current = self.current_cell

        if current is not None and current.cell_type == cell_type == "markdown":
            return

        types = {"markdown": nbformat.new_markdown_cell, "code": nbformat.new_code_cell}
        new_cell = types[cell_type]
        self.cells.append(new_cell())

    def append(self, text: str):

        if self.prefix is not None:
            text = text.replace("\n", f"\n{self.prefix}")

        self.current_cell.source += text

    # --------------------------------- Visitors --------------------------------------
    visit_compound, depart_compound = no_op
    visit_compact_paragraph, depart_compact_paragraph = no_op
    visit_document, depart_document = no_op
    visit_figure, depart_figure = no_op
    visit_legend, depart_legend = no_op
    visit_solution, depart_solution = no_op
    visit_target, depart_target = no_op
    visit_tutorial, depart_tutorial = no_op

    def visit_caption(self, node):
        self.current_cell.source += "*"

    def depart_caption(self, node):
        self.current_cell.source += "*\n"

    def visit_bullet_list(self, node):
        self.new_cell("markdown")
        self.current_cell.source += "\n"

    def depart_bullet_list(self, node):
        self.current_cell.source += "\n"

    def visit_comment(self, node):
        self.new_cell("markdown")

    def depart_comment(self, node):
        pass

    def visit_emphasis(self, node):
        self.current_cell.source += "*"

    def depart_emphasis(self, node):
        self.current_cell.source += "*"

    def visit_image(self, node):
        self.new_cell("markdown")

        uri = pathlib.Path(node["uri"]).name
        fpath = pathlib.Path(RESOURCES, uri)
        self.current_cell.source += f"\n![]({fpath})\n"

    def depart_image(self, node):
        pass

    def visit_inline(self, node):
        self.current_cell.source += "["

    def depart_inline(self, node):
        self.current_cell.source += "]"

    def visit_list_item(self, node):
        self.append("- ")

    def depart_list_item(self, node):
        pass

    def visit_literal(self, node):
        self.current_cell.source += "`"

    def depart_literal(self, node):
        self.current_cell.source += "`"

    def visit_literal_block(self, node):
        self.new_cell("code")

    def depart_literal_block(self, node):
        pass

    def visit_note(self, node):
        self.new_cell("markdown")
        self.append("> **Note**\n")
        self.prefix = "> "

    def depart_note(self, node):
        self.prefix = None

    def visit_paragraph(self, node):
        self.new_cell("markdown")

        if isinstance(node.parent, nodes.list_item):
            return

        self.append("\n")

    def depart_paragraph(self, node):
        self.append("\n")

    def visit_reference(self, node):
        self.current_cell.source += "["

    def depart_reference(self, node):
        url = node.attributes["refuri"]
        self.current_cell.source += f"]({url})"

    def visit_section(self, node):
        self.level += 1
        self.new_cell("markdown")

    def depart_section(self, node):
        self.level -= 1

    def visit_strong(self, node):
        self.current_cell.source += "**"

    def depart_strong(self, node):
        self.current_cell.source += "**"

    def visit_Text(self, node):

        # Don't emit anything for rst comments.
        if isinstance(node.parent, nodes.comment):
            return

        # Nothing special to do for markdown cells.
        if self.current_cell.cell_type == "markdown":
            self.append(node.astext())
            return

        # If we're processing code, then strip any doctest markers
        pattern = re.compile("^(>>>|\\.\\.\\.) ?")
        source = "\n".join(
            [pattern.sub("", line) for line in node.astext().split("\n")]
        )
        self.append(source)

    def depart_Text(self, node):
        pass

    def visit_title(self, node):
        title = "#" * self.level
        self.append(f"{title} ")

    def depart_title(self, node):
        self.append("\n")


def setup(app):
    app.add_node(solution, html=(visit_solution, depart_solution))
    app.add_node(tutorial, html=(visit_tutorial, depart_tutorial))

    app.add_directive(SolutionDirective.NAME, SolutionDirective)
    app.add_directive(TutorialDirective.NAME, TutorialDirective)

    return {"version": esbonio.__version__, "parallel_read_safe": True}



import argparse
import logging
import esbonio.lsp as lsp

cli = argparse.ArgumentParser(prog="esbonio", description="Your one true lsp.")

logging.basicConfig(
    level=logging.DEBUG,
    format="[%(levelname)s][%(name)s]: %(message)s",
    filemode="w",
    filename="/home/alex/Projects/esbonio/lib/esbonio/lsp.log",
)

lsp.server.start_io()



import inspect
import importlib
import logging
import pathlib
import re
from typing import Dict, Optional

import appdirs
import docutils.parsers.rst.directives as directives
import docutils.parsers.rst.roles as roles

from pygls.features import COMPLETION, INITIALIZED
from pygls.server import LanguageServer
from pygls.types import (
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionParams,
    MessageType,
)

from sphinx.application import Sphinx


def completion_from_directive(name, directive) -> CompletionItem:
    """Convert an rst directive to a completion item we can return to the client."""

    # 'Core' docutils directives are listed as tuples (modulename, ClassName) so we
    # have to go and look them up ourselves.
    if isinstance(directive, tuple):
        mod, cls = directive
        module = importlib.import_module(f"docutils.parsers.rst.directives.{mod}")
        directive = getattr(module, cls)

    documentation = inspect.getdoc(directive)
    return CompletionItem(
        name,
        kind=CompletionItemKind.Class,
        detail="directive",
        documentation=documentation,
    )


def completion_from_role(name, role) -> CompletionItem:
    """Convert an rst directive to a completion item we can return to the client."""
    return CompletionItem(name, kind=CompletionItemKind.Function, detail="role")


class RstLanguageServer(LanguageServer):
    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(__name__)

        self.app = None
        """Sphinx application instance configured for the current project."""

        self.directives: Optional[Dict[str, CompletionItem]] = None
        """Dictionary holding the directives that have been registered."""

        self.roles: Optional[Dict[str, CompletionItem]] = None
        """Dictionary holding the roles that have been registered."""


server = RstLanguageServer()


@server.feature(INITIALIZED)
def on_initialized(rst: RstLanguageServer, params):
    """Do set up once the initial handshake has been completed."""
    rst.logger.debug("Workspace root %s", rst.workspace.root_uri)

    root = pathlib.Path(rst.workspace.root_uri.replace("file://", ""))
    candidates = list(root.glob("**/conf.py"))

    if len(candidates) == 0:
        rst.show_message(
            "Unable to find your 'conf.py', features will be limited",
            msg_type=MessageType.Warning,
        )

    else:
        # TODO: #1 Multi root workspaces?
        # TODO: Multi sphinx projects?
        src = candidates[0].parent
        rst.logger.debug("Found config dir %s", src)
        build = appdirs.user_cache_dir("esbonio", "swyddfa")
        rst.app = Sphinx(src, src, build, build, "html", status=None, warning=None)

    # Lookup the directives and roles that have been registered
    dirs = {**directives._directive_registry, **directives._directives}
    rst.directives = {k: completion_from_directive(k, v) for k, v in dirs.items()}

    role_s = {**roles._roles, **roles._role_registry}
    rst.roles = {k: completion_from_role(k, v) for k, v in role_s.items()}


import re

NEW_DIRECTIVE = re.compile("\\s*\\.\\.\\s+([\\w-]+)?")
NEW_ROLE = re.compile("(^|\\s+):([\\w-]+)?")


@server.feature(COMPLETION, trigger_characters=["."])
def completions(rst: RstLanguageServer, params: CompletionParams):
    uri = params.textDocument.uri
    pos = params.position

    doc = rst.workspace.get_document(uri)
    line = doc.lines[pos.line]

    if NEW_DIRECTIVE.match(line):
        candidates = list(rst.directives.values())

    elif NEW_ROLE.match(line):
        candidates = list(rst.roles.values())

    else:
        candidates = [*rst.directives.values(), *rst.roles.values()]

    return CompletionList(False, candidates)
